#!/usr/bin/env node
/*jslint nomen: true, node: true, todo: true*/
"use strict";
// gamelab by Seandon Mooy 2013 -> http://github.com/erulabs/gamelab
(function () {
    var version = '0.1',
        metaFile = '.gamelab',
        fs = require('fs'),
        os = require('os'),
        childPrc = require('child_process'),
        exec = childPrc.exec,
        spawn = childPrc.spawn,
        program = require('commander'),
        coffee = require('coffee-script'),
        ncp = require('ncp').ncp,
        // Remove request dep... Can easily just use http.
        request = require('request'),
        watch = require('node-watch'),
        sass = require('node-sass'),
        cwd = process.cwd(),
        processHandle = false,
        colors = { red: "\u001b[31m", blue: "\u001b[34m", green: "\u001b[32m", cyan: "\u001b[36m", reset: "\u001b[0m" },
        projectName,
        meta = {
            deliniator: '/'
        },
        gamelabMetaString,
        serverTarget;
    if (os.platform() === "win32") {
        meta.deliniator = "\\";
    }
    // Todo: make a CLI argument for what server filename to wrap.
    serverTarget = meta.deliniator + 'app' + meta.deliniator + 'core.js';
    // Run functor(item, callback) on a list of items.
    function walkArray(list, index, direction, functor, callback) {
        var destination = list.length;
        if (direction === -1) {
            destination = 0;
        }
        if (index !== destination) {
            functor(list[index], index + direction, function () {
                walkArray(list, index + direction, direction, functor, callback);
            });
        } else {
            callback();
        }
    }
    // Simple exit
    function exit(message) {
        console.log(colors.red, message, colors.reset);
        process.exit();
    }
    // Sync read any given JSON file - if errorMessage is defined it will error out with that if not found, if not, returns false;
    function readJSON(file, errorMessage, callback) {
        fs.exists(file, function (exists) {
            if (exists) {
                fs.readFile(file, 'utf8', function (error, data) {
                    callback(error, JSON.parse(data));
                });
            } else if (errorMessage) {
                return exit(errorMessage);
            }
        });
    }
    // Recursivly search up the directory path, looking for a .gamelab file - the directory it resides in is the current project
    function gamelabMeta(callback) {
        var thisMetaTry,
            path;
        path = (cwd + meta.deliniator).split(meta.deliniator);
        walkArray(path, path.length - 1, -1, function (item, index, proceed) {
            if (item === "" && index === path.length - 2) {
                thisMetaTry = '.';
            } else {
                thisMetaTry = path.slice(0, path.indexOf(item) - path.length).join(meta.deliniator);
            }
            fs.exists(thisMetaTry + meta.deliniator + metaFile, function (exists) {
                if (exists) {
                    readJSON(thisMetaTry + meta.deliniator + metaFile, false, function (error, data) {
                        if (error === undefined) {
                            if (error) {
                                console.log('readJSON error:', error);
                            }
                        }
                        var deliniator = meta.deliniator;
                        meta = data;
                        meta.deliniator = deliniator;
                        meta.projectPath = thisMetaTry;
                        //TODO: This doesn't work, however, the idea is clear and it -ought- to work.
                        //exec('export NODE_ENV='+meta.mode);
                        callback();
                    });
                } else {
                    proceed();
                }
            });
        }, function () {
            console.log('Not a valid gamelab directory');
        });
    }
    // Compile coffeescript file into javascript
    function coffee2js(file, callback) {
        var compileStart = new Date(),
            shortFileName = file,
            compileJavascript,
            compileTarget,
            compileBreakout,
            compileFinish;
        fs.readFile(file, 'utf8', function (err, coffeeCode) {
            var i;
            if (err) {
                throw err;
            }
            try {
                compileJavascript = coffee.compile(coffeeCode);
            } catch (error) {
                compileBreakout = coffeeCode.split("\n");
                console.log(colors.red, 'Compile error', colors.reset, 'in', shortFileName, 'line #' + error.location.first_line + ':');
                for (i = error.location.first_line - 2; i < error.location.first_line + 2; i += 1) {
                    console.log('#' + i, compileBreakout[i]);
                }
            } finally {
                compileTarget = file.replace('.coffee', '.js');
                fs.writeFile(compileTarget, compileJavascript, function () {
                    compileFinish = new Date();
                    console.log('Compiled', shortFileName, 'in', (compileFinish - compileStart) / 1000, 'seconds');
                    if (typeof callback === "function") {
                        callback();
                    }
                });
            }
        });
    }
    // Compile SCSS into CSS
    function sass2css(file, callback) {
        var compileStart = new Date(),
            shortFileName = file,
            compileTarget,
            compileFinish;
        compileTarget = file.replace('.scss', '.css');
        sass.render({
            file: file,
            success: function (css) {
                fs.writeFile(compileTarget, css, function () {
                    compileFinish = new Date();
                    console.log('Compiled', shortFileName, 'in', (compileFinish - compileStart) / 1000, 'seconds');
                    if (typeof callback === "function") {
                        callback();
                    }
                });
            },
            error: function (error) {
                console.log('SCSS Compile error:', error);
            }
        });
    }
    // Start the server (essentially wrap "node SERVERFILE")
    function startServer() {
        var serverTargetFinal = meta.projectPath + serverTarget;
        fs.exists(serverTargetFinal, function (exists) {
            if (exists) {
                processHandle = spawn("node", [meta.projectPath + serverTarget]);
                //processHandle.on('close', function (code, signal) {
                //    // Do nothing
                //});
                processHandle.stdout.on('data', function (data) {
                    process.stdout.write(data.toString());
                });
                processHandle.stderr.on('data', function (data) {
                    process.stdout.write(data.toString());
                });
            } else {
                console.log('Unable to locate', serverTargetFinal);
            }
        });
    }
    // Stop the server
    function stopServer() {
        //console.log(colors.red, 'Server stopped', colors.reset);
        processHandle.kill();
    }
    // Restart the server
    function restartServer() {
        console.log(colors.green, 'Server refreshed', colors.reset);
        stopServer();
        startServer();
    }
    // Watch a directory and send file updates to the server if server == true
    function watchDirectory() {
        if (meta.mode !== 'development') {
            exit('Development server cannot be started in production mode');
        }
        watch([ // TODO: Watch files listed in .gamelab
            meta.projectPath + '/app/www/assets',
            meta.projectPath + '/app/www/templates',
            meta.projectPath + '/app/www/index.ejs',
            meta.projectPath + '/app/core.js',
            meta.projectPath + '/app/app.js'
        ], function (file) {
            if (file.indexOf('.coffee') !== -1) {
                coffee2js(file);
                return false;
            }
            if (file.indexOf('.scss') !== -1) {
                sass2css(file);
                return false;
            }
            process.stdout.write('file "' + file + '" changed:');
            request.post('http://localhost:9000/__quickReload', {
                form: { file: file }
            }, function (error, response) {
                if (error) {
                    console.log('Quick reload error: ', error);
                }
                if (response.body === "server") {
                    restartServer();
                } else {
                    console.log(colors.green, 'Client refreshed', colors.reset);
                }
            });
        });
        startServer();
    }
    // Save out the meta information to the projects .gamelab file
    function writeMetaString(projectPath) {
        var osSafeDirname = __dirname;
        if (os.platform() === "win32") {
            osSafeDirname = __dirname.replace(/\\/g, "\\\\");
        }
        // Write metadata template file - Make sure this is always pretty-printed human-readable valid JSON
        gamelabMetaString = "{\n\t\"path\":\t\"" + osSafeDirname + "\",\n\t\"mode\":\t\"development\"\n}";
        fs.writeFile(projectPath + meta.deliniator + metaFile, gamelabMetaString, function (err) {
            if (err) {
                return console.error(err);
            }
        });
    }
    function clientStep(callback) {
        exec('bower install', function (stdout, stderr, error) {
            if (stdout) {
                console.log(stdout);
            }
            if (stderr) {
                console.log(stderr);
            }
            if (error) {
                console.log(colors.red, error, colors.reset);
            }
            callback();
        });
    }
    function serverStep(callback) {
        exec('npm install', function (stdout, stderr, error) {
            if (stdout) {
                console.log(stdout);
            }
            if (stderr) {
                console.log(stderr);
            }
            if (error) {
                console.log(colors.red, error, colors.reset);
            }
            callback();
        });
    }
    function installComplete() {
        console.log('install complete!');
    }
    // Usage / help (commander.js)
    program.version(version)
        .option('init [projectName]', 'New Project')
        .option('install', 'Install a projects node modules and static assets')
        //.option('createClass [className] <type>', 'Create a template object - <type> one of client/server/shared/all defaults to all')
        //.option('list', 'List available gamelab modules')
        .option('server', 'Start development server')
        //.option('coffee', 'Watch for changes to and compile Coffee-Script files')
        //.option('build', 'Compile Project for production mode')
        //.option('lint', 'Lint project files')
        .parse(process.argv);
    if (program.init) {
        projectName = program.init; //TODO: Validate the name of this
        // Recursivly copy template project into place
        ncp(__dirname + meta.deliniator + 'templateProject', projectName, function (err) {
            if (err) {
                return console.error(err);
            }
            writeMetaString(cwd + meta.deliniator + projectName);
        });
    } else if (program.build) {
        // TODO: Minify project
        gamelabMeta(function () {
            console.log('Unimplimented');
        });
    } else if (program.createClass) {
        console.log('createclass');
    } else if (program.list) {
        console.log('lint');
    } else if (program.server) {
        // Start server (init.js)
        gamelabMeta(function () {
            watchDirectory(true);
        });
    } else if (program.coffee) {
        gamelabMeta(function () {
            watchDirectory(false);
        });
    } else if (program.install) {
        gamelabMeta(function () {
            serverStep(function () {
                clientStep(function () {
                    installComplete();
                });
            });
        });
    } else if (program.lint) {
        // TODO: Lint project
        gamelabMeta();
    } else {
        program.help();
    }
}());
